/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020                                    ***
***  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_ACCOUNTING_ACCOUNTCODE_H_
#define DRN_ACCOUNTING_ACCOUNTCODE_H_

#include <QtGlobal>

#include <cstdint>
#include <iosfwd>
#include <string>

#include "accounting_export.h"
#include <foundation/Optional.hpp>


namespace drn
{
namespace accounting
{
namespace tests
{

struct AccountNumberUT;

}

/**
 * @brief The representation of the unique identifier for an account.
 */
class DRN_ACCOUNTING_EXPORT AccountNumber
{
	friend tests::AccountNumberUT;

	/// The number that is an invalid for an account code.
	static const std::uint32_t invalid_;
	std::uint32_t value_;

public:
	using ValueType = std::uint32_t;

	/**
	 * @brief Default initialises the account number to an invalid state.
	 */
	AccountNumber();
	/**
	 * @brief Fully initialises the account number.
	 *
	 * @exception Error When the value is not set.
	 *
	 * @param value The value for the account.
	 */
	explicit AccountNumber(ValueType value);
	AccountNumber(const AccountNumber&) = default;
	AccountNumber(AccountNumber&&) noexcept = default;
	AccountNumber& operator=(const AccountNumber&) = default;
	AccountNumber& operator=(AccountNumber&&) noexcept = default;
	/**
	 * @brief The underlying value of the account number.
	 */
	const ValueType& value() const;
	/**
	 * @brief Changes the value for the account number.
	 *
	 * @exception Error When the value is not set.
	 *
	 * @param value The new value for the account.
	 */
	void setValue(ValueType value);
	/**
	 * @brief Determines if the account number is set.
	 *
	 * @return Gives true when the value is set, false otherwise.
	 */
	bool isEmpty() const;
};
/**
 * @brief The representation of a unique identifier and its associated name for an account. The only
 * identifying item is the number, the name is a convenience for the user to associate the number
 * with a name.
 */
class DRN_ACCOUNTING_EXPORT AccountCode
{
	AccountNumber number_;
	std::string name_;
	foundation::Optional<AccountNumber> parent_;

public:
	AccountCode() = default;
	/**
	 * @brief Partially initialises the object with supplied account code name, and defaults the
	 * rest.
	 *
	 * @exception AccountError When the supplied name is effectively empty.
	 * @exception AccountError When the supplied name contains the account separator.
	 * @exception AccountError When the parent is invalid.
	 *
	 * @param name The unique name for the account, e.g. Cash.
	 * @param parent (Optional) The parent account number for the account.
	 */
	explicit AccountCode(const std::string& name, foundation::Optional<AccountNumber> parent = {});
	/**
	 * @brief Partially initialises the object with supplied values, and defaults the rest.
	 *
	 * @exception AccountError When the number is not set.
	 * @exception AccountError When the parent and the value are the same.
	 * @exception AccountError When the parent is invalid.
	 *
	 * @param number The unique number for the account, e.g. 1000.
	 * @param parent (Optional) The parent account number for the account.
	 */
	explicit AccountCode(AccountNumber number, foundation::Optional<AccountNumber> parent = {});
	/**
	 * @brief Fully initialises the object.
	 *
	 * @exception AccountError When the supplied number is zero.
	 * @exception AccountError When the supplied name is effectively empty.
	 * @exception AccountError When the supplied name contains the account separator.
	 *
	 * @param number The unique number for the account, e.g. 1000.
	 * @param name The unique name for the account, e.g. Cash.
	 * @param parent (Optional) The parent account number for the account.
	 */
	AccountCode(
		const AccountNumber number,
		const std::string& name,
		foundation::Optional<AccountNumber> parent = {}
	);
	AccountCode(const AccountCode&) = default;
	AccountCode(AccountCode&&) = default;
	~AccountCode() = default;
	AccountCode& operator=(const AccountCode&) = default;
	AccountCode& operator=(AccountCode&&) = default;
	/**
	 * @brief Less-than operator compares the unique identifier number.
	 *
	 * @param other The other object whose number should be greater than the current object.
	 *
	 * @return When the current object's number is less-than the supplied object it'll give true,
	 * else false.
	 */
	bool operator<(const AccountCode& other) const;
	/**
	 * @brief Equality operator compares the unique identifier number.
	 *
	 * @param other The other object whose unique identifier number should be equal.
	 *
	 * @return When the unique identifier numbers are the same it'll give true, else false.
	 */
	bool operator==(const AccountCode& other) const;
	/**
	 * @brief Inequality operator compares the unique identifier number.
	 *
	 * @param other The other object whose unique identifier number should not be equal.
	 *
	 * @return When the unique identifier numbers are not the same it'll give true, else false.
	 */
	bool operator!=(const AccountCode& other) const;
	/**
	 * @brief The numeric identifier for the account.
	 */
	const AccountNumber& number() const;
	/**
	 * @brief Changes the number for the account.
	 *
	 * @exception Error When the parent and the number are the same.
	 *
	 * @param number The new number for the account.
	 */
	void setNumber(AccountNumber number);
	/**
	 * @brief The human friendly name for the account.
	 */
	const std::string& name() const;
	/**
	 * @brief Changes the human friendly name for the account.
	 *
	 * @exception AccountError When the supplied identifier is effectively empty.
	 *
	 * @param name The new name for the account.
	 */
	void setName(std::string name);
	/**
	 * @brief Determines if the account code unique identifier is empty. e.g. default constructed.
	 */
	bool isEmpty() const;
	/**
	 * @brief The number of the parent account for this account.
	 */
	const foundation::Optional<AccountNumber>& parent() const;
	/**
	 * @brief Changes the parent for the account.
	 *
	 * @exception Error When the parent is invalid.
	 * @exception Error When the parent and the current number are the same.
	 *
	 * @param parent The new parent for the account.
	 */
	void setParent(foundation::Optional<AccountNumber> parent);

	/// The character separator to place between the number and name when outputting the account
	/// code, currently an em-dash.
	static const std::string separator_;
	/// The number of characters that the account code number should always be, currently four.
	static const std::uint8_t fieldWidth_;
};

DRN_ACCOUNTING_EXPORT bool operator<(const AccountNumber& lhs, const AccountNumber& rhs);
DRN_ACCOUNTING_EXPORT bool operator>(const AccountNumber& lhs, const AccountNumber& rhs);
DRN_ACCOUNTING_EXPORT bool operator==(const AccountNumber& lhs, const AccountNumber& rhs);
DRN_ACCOUNTING_EXPORT bool operator!=(const AccountNumber& lhs, const AccountNumber& rhs);
DRN_ACCOUNTING_EXPORT bool operator==(const AccountCode& lhs, const AccountNumber& rhs);
DRN_ACCOUNTING_EXPORT bool operator!=(const AccountCode& lhs, const AccountNumber& rhs);
DRN_ACCOUNTING_EXPORT bool operator==(const AccountNumber& lhs, const AccountCode& rhs);
DRN_ACCOUNTING_EXPORT bool operator!=(const AccountNumber& lhs, const AccountCode& rhs);
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const AccountNumber& an);
DRN_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const AccountCode& ac);
DRN_ACCOUNTING_EXPORT std::string toHumanReadable(const AccountNumber& an);
/**
 * @brief Converts an account code into a human readable form.
 *
 * @param ac The account code to convert.
 *
 * @return The account code in the form: NumberSeparatorName.
 */
DRN_ACCOUNTING_EXPORT std::string toHumanReadable(const AccountCode& ac);
DRN_ACCOUNTING_EXPORT uint qHash(const AccountCode& ac);

}}

#endif
