/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020                                    ***
***  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGET_ACCOUNTING_BANK_H_
#define DRN_BUDGET_ACCOUNTING_BANK_H_

#include <iosfwd>
#include <set>
#include <string>

#include <accounting/AccountCode.h>
#include "budget-accounting_export.h"


namespace drn
{
namespace budget_accounting
{

/**
 * @brief A representation of the name of a financial institution.
 */
class DRN_BUDGET_ACCOUNTING_EXPORT BankName
{
	std::string value_;

public:
	BankName() = default;
	/**
	 * @brief The unique name of the financial institution.
	 *
	 * @exception BankError When the supplied name is effectively empty.
	 *
	 * @param name The name of the financial institution.
	 */
	explicit BankName(std::string name);
	BankName(const BankName&) = default;
	BankName(BankName&&) noexcept = default;
	~BankName() = default;
	BankName& operator=(const BankName&) = default;
	BankName& operator=(BankName&&) noexcept = default;
	bool operator<(const BankName& other) const;
	bool operator==(const BankName& other) const;
	bool operator!=(const BankName& other) const;
	/**
	 * @brief An accessor to the value for the name of the financial institution.
	 */
	const std::string& value() const noexcept;
	/**
	 * @brief Determines if the financial institution's name value is set.
	 *
	 * @return True when the name value is not set, false otherwise.
	 */
	bool isEmpty() const;
};
/**
 * @brief A representation of a financial institution.
 */
class DRN_BUDGET_ACCOUNTING_EXPORT Bank
{
	std::set<accounting::AccountNumber> accountNumbers_;
	bool isClosed_;

public:
	/**
	 * @brief Default constructor initialising all members to empty.
	 */
	Bank() = default;
	/**
	 * @brief Full and partial initialising constructor.
	 *
	 * @exception BankError When the supplied name is empty.
	 *
	 * @param name The name of the financial institution.
	 * @param numbers The general ledger account numbers associated with the financial institution.
	 * @param isClosed The flag to indicate if a financial institution is closed for usage.
	 */
	explicit Bank(
		BankName name,
		std::set<accounting::AccountNumber> numbers = {},
		const bool isClosed = false
	);
	/**
	 * @brief An accessor for all the accounts associated with this financial institution.
	 *
	 * @return The account numbers of the associated accounts, empty when not set.
	 */
	const std::set<accounting::AccountNumber>& accountNumbers() const;
	/**
	* @brief An accessor for the flag to indicate if the financial institution is closed for usage.
	*
	* @return Gives true when the financial institution is closed.
	*/
	const bool& isClosed() const;
	/**
	 * @brief Marks the financial institution as closed for usage.
	 *
	 * @exception BankError When the financial institution is already closed.
	 */
	void close();
	/**
	 * @brief Marks the financial institution as open for usage.
	 *
	 * @exception BankError When the financial institution is already opened.
	 */
	void open();
	/**
	 * @brief Adds an account number to associate with the financial institution.
	 *
	 * @param number The account number to add.
	 */
	void add(accounting::AccountNumber number);
	/**
	 * @brief Removes an account number from the association with the financial institution.
	 *
	 * @exception BankError When the account number to remove is not associated with the financial
	 * institution.
	 *
	 * @param number The number to remove.
	 */
	void remove(const accounting::AccountNumber& number);

	/// The name of the financial institution, empty when not set.
	BankName name_;
};

DRN_BUDGET_ACCOUNTING_EXPORT std::string operator+(const std::string& value, const BankName& bn);
DRN_BUDGET_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const BankName& bn);
DRN_BUDGET_ACCOUNTING_EXPORT bool operator==(const Bank& lhs, const Bank& rhs);
DRN_BUDGET_ACCOUNTING_EXPORT bool operator!=(const Bank& lhs, const Bank& rhs);
DRN_BUDGET_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const Bank& b);
DRN_BUDGET_ACCOUNTING_EXPORT uint qHash(const BankName& name);

}}

#endif
