/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020                                    ***
***  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGET_ACCOUNTING_BANKLEDGERS_H_
#define DRN_BUDGET_ACCOUNTING_BANKLEDGERS_H_

#include <array>
#include <cstdint>
#include <iosfwd>
#include <map>
#include <string>

#include "budget-accounting_export.h"

#include <accounting/AccountCode.h>
#include "Bank.h"
#include "foundation/ObserverPtr.hpp"


template<typename>
class QSet;

namespace pecunia
{
namespace currency
{

class Money;

}}

namespace drn
{
namespace accounting
{

class AccountCode;
class AccountCodeNumberGenerator;
enum class AccountTypes : std::uint8_t;
class GeneralLedger;

}
namespace foundation
{

template<typename>
class Optional;

}
namespace budget_accounting
{

struct BankMap;
class BankName;

/**
 * @brief The types of financial accounts supported.
 *
 */
enum class DRN_BUDGET_ACCOUNTING_EXPORT BankAccountTypes
{
	/// The bank account type is not known.
	Unknown,
	/// The bank account type is a for chequeing/current account.
	Chequeing,
	/// The bank account type is a credit card.
	CreditCard,
	/// The bank account type is for a loan.
	Loan,
	/// The bank account type is an interest bearing savings account.
	Savings,
};
/**
 * @brief A manager for the relationship between the general ledger and the banks.
 */
class DRN_BUDGET_ACCOUNTING_EXPORT BankLedgers
{
	foundation::ObserverPtr<BankMap> banks_;
	foundation::ObserverPtr<accounting::GeneralLedger> ledgers_;
	std::map<BankName, std::map<accounting::AccountNumber, BankAccountTypes>> bankAccountTypes_;

public:
	/**
	 * @brief Fully constructs the associations for banks and the general ledger.
	 *
	 * @exception BankError When a financial institution uses the accounts of another financial
	 * institution.
	 * @exception BankError When a financial institution user understandable account type is not
	 * known.
	 * @exception BankError When a financial institution has no associations.
	 * @exception BankError When a financial institution uses the a default ledger account.
	 *
	 * @param banks The banks to be managed.
	 * @param ledgers The ledgers for the banks to use.
	 */
	BankLedgers(
		foundation::ObserverPtr<BankMap> banks,
		foundation::ObserverPtr<accounting::GeneralLedger> ledgers,
		std::map<BankName, std::map<accounting::AccountNumber, BankAccountTypes>> bankAccountTypes
	);
	/**
	 * @brief An accessor for the banks being managed.
	 *
	 * @return The current banks managed.
	 */
	const BankMap& banks() const;
	/**
	 * @brief An accessor for the ledgers being managed.
	 *
	 * @return The current ledgers managed.
	 */
	const accounting::GeneralLedger& ledgers() const;
	/**
	 * @brief An accessor to the association between a bank, its accounts and the user
	 * understandable bank account type.
	 *
	 * @return The current association of bank and their account types..
	 */
	const std::map<
		BankName,
		std::map<accounting::AccountNumber, BankAccountTypes>
	>& bankAccountTypes() const;
	/**
	 * @brief Locates the name of a financial institution associated with a general ledger account.
	 *
	 * @param number The account number to find the associated financial institution.
	 *
	 * @return When found gives the financial institution identifier, else an empty value.
	 */
	foundation::Optional<BankName> findBankName(const accounting::AccountNumber& number);
	/**
	 * @brief Adds a bank to the system. Any accounts associated with this bank will be associated
	 * as an unknown bank account type.
	 *
	 * @exception BankError When the accounts associated with the bank are not in the general
	 * ledger.
	 * @exception BankError When the associated account ledger in the financial institution does not
	 * match the associations with the bank types.
	 *
	 * @param bank The bank to be added.
	 * @param accountNumbersBankTypes The account ledger associations.
	 */
	void add(
		Bank bank,
		std::map<accounting::AccountNumber, BankAccountTypes> accountNumbersBankTypes = {}
	);
	/**
	 * @brief Adds an account to the system.
	 *
	 * @exception BankError When the bank does not exist.
	 *
	 * @param bn The bank to associate the account with.
	 * @param accountName The name of the account.
	 * @param type The type of bank account that is associated.
	 * @param balance The opening balance of the account.
	 * @param generator The account code number generator ready to be used.
	 */
	void addAccount(
		const BankName& bn,
		const std::string& accountName,
		const BankAccountTypes& type,
		const pecunia::currency::Money& balance,
		accounting::AccountCodeNumberGenerator& generator
	);
	/**
	 * @brief Adds an association with an account number to a bank.
	 *
	 * @exception BankError When the bank does not exist.
	 * @exception BankError When another bank is already associated with that number.
	 * @exception BankError When the account number does not exist.
	 *
	 * @param name The name of the bank to add the account number to.
	 * @param number The number to add to the bank.
	 * @param type The type of bank account the general ledger account is managing.
	 */
	void addBankAccount(
		const BankName& name,
		const accounting::AccountNumber& number,
		const budget_accounting::BankAccountTypes& type
	);
	/**
	 * @brief Renames the financial institution.
	 *
	 * @exception BankError When the financial institution name to be changed does not exist.
	 * @exception BankError When the financial institution name to be changed and to be changed
	 * into are the same.
	 * @exception BankError When the financial institution name to change to already exists.
	 * @exception BankError When the financial institution is already closed.
	 *
	 * @param original The current name of the financial institution.
	 * @param changed The changed name of the financial institution.
	 */
	void rename(const BankName& original, const BankName& changed);
	/**
	 * @brief Marks the financial institution and associated accounts as closed for usage.
	 *
	 * @exception BankError When the financial institution does not exist.
	 *
	 * @param name The name of the financial institution.
	 */
	void close(const BankName& name);
	/**
	 * @brief Marks the financial institution and associated accounts as opened for usage.
	 *
	 * @exception BankError When the financial institution does not exist.
	 *
	 * @param name The name of the financial institution.
	 */
	void open(const BankName& name);
	/**
	 * @brief Determines if the container of bank accounts is empty.
	 *
	 * @return Gives true when there are not any banks in the container, false otherwise.
	 */
	bool isEmpty() const;
	/**
	 * @brief Clears all the bank account types associated with banks.
	 */
	void clearBankAccountType();
	/**
	 * @brief Removes a financial institution and all ledger accounts associated with it.
	 *
	 * @exception BankError When the financial institution does not exist.
	 * @exception BankError When the financial institution is closed.
	 * @exception AccountError When any of the financial institution's associated ledger accounts
	 * cannot be removed.
	 *
	 * @param name The name of the financial institution to remove.
	 */
	void remove(const BankName& name);
	/**
	 * @brief Updates the financial institution information and the account ledger associations.
	 *
	 * @exception BankError When the financial information to update is closed.
	 * @exception BankError When the associated account ledger in the financial institution does not
	 * match the associations with the bank types.
	 * @exception BankError When the account ledger to associate does not exist.
	 *
	 * @param bank The financial institution to update.
	 * @param accountNumbersBankTypes The new account ledger associations.
	 */
	void update(
		Bank bank,
		const std::map<accounting::AccountNumber, BankAccountTypes>& accountNumbersBankTypes
	);
	/**
	 * @brief Recodes the financial institution account association.
	 *
	 * @exception AccountError When no financial institution is associated with the original
	 * account.
	 * @exception BankError When a financial institution is already associated with the new
	 * account.
	 * @exception BankError When the association removal fails.
	 *
	 * @param original The original number for the financial institution account association.
	 * @param changed The new number for the financial institution account association.
	 */
	void recode(
		const accounting::AccountNumber& original,
		const accounting::AccountNumber& changed
	);
	/**
	 * @brief Removes an account association from a financial institution.
	 *
	 * @exception BankError When the financial institution does not exist.
	 * @exception BankError When no financial institution is associated with the original account.
	 *
	 * @param name The financial institution to remove the account association from.
	 * @param number The account number to be removed.
	 */
	void removeBankAccount(const BankName& name, const accounting::AccountNumber& number);
	/**
	 * @brief Looks-up all account codes associated with a given bank.
	 *
	 * @param name The name of the bank whose account codes are desired.
	 *
	 * @return The account codes associated with the bank or an empty container when either the bank
	 * does not exist or has no associated ledger accounts.
	 */
	QSet<accounting::AccountCode> bankAccountCodes(const BankName& name) const;
};

DRN_BUDGET_ACCOUNTING_EXPORT BankAccountTypes toBankAccountTypes(const std::string& bankAccountType);
DRN_BUDGET_ACCOUNTING_EXPORT std::string toHumanReadable(const BankAccountTypes& bat);
DRN_BUDGET_ACCOUNTING_EXPORT accounting::AccountTypes toAccountTypes(const BankAccountTypes& type);
DRN_BUDGET_ACCOUNTING_EXPORT bool operator==(const BankLedgers& lhs, const BankLedgers& rhs);
DRN_BUDGET_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const BankAccountTypes& bat);
DRN_BUDGET_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const BankLedgers& ba);

constexpr std::uint8_t numberOfBankAccountTypes{5};
extern DRN_BUDGET_ACCOUNTING_EXPORT const std::array<BankAccountTypes, numberOfBankAccountTypes> allBankAccountTypes;

}}

#endif
