/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020                                    ***
***  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGET_ACCOUNTING_BUDGETBANKLEDGERS_H_
#define DRN_BUDGET_ACCOUNTING_BUDGETBANKLEDGERS_H_

#include <cassert>
#include <iosfwd>
#include <map>
#include <memory>
#include <string>
#include <type_traits>
#include <utility>

#include <accounting/LedgerBooks.h>
#include "BankLedgers.h"
#include "BankMap.h"
#include <budgeting/Budget.h>
#include "BudgetLedgers.h"
#include "budget-accounting_export.h"
#include <foundation/TypeIndexMap.hpp>


template<typename>
class QSet;

namespace date
{

class year_month_day;

}

namespace pecunia
{
namespace currency
{

enum class Codes;

}}

namespace drn
{
namespace accounting
{

class Account;
class AccountCode;
class AccountCodeNumberGenerator;
class Transaction;
class TransactionNumber;

}
namespace budgeting
{

struct BudgetItemDue;
class BudgetSource;

}
namespace budget_accounting
{

class Bank;
class BankName;

/**
 * @brief Manages the relationship between the budget, the general ledger, and bank accounts.
 */
class DRN_BUDGET_ACCOUNTING_EXPORT BudgetBankLedgers
{
	void verifyBudgetSourceName(const budgeting::BudgetSource& bs);
	void verifyAccountCodeName(const accounting::AccountCode& ac);

	accounting::GeneralLedger ledgers_;
	budgeting::Budget budget_;
	BankMap banks_;
	std::unique_ptr<accounting::AccountCodeNumberGenerator> accountCodeNumberGenerator_;
	BudgetLedgers budgetLedgers_;
	BankLedgers bankLedgers_;

public:
	BudgetBankLedgers();
	/**
	 * @brief Fully constructs the object.
	 *
	 * @pre The accountCodeNumberGenerator cannot be null.
	 *
	 * @exception AccountError When the ledger contains an account code name that is same as a
	 * budget item source.
	 *
	 * @param ledgers The values for the General Ledger.
	 * @param budget The values for the budget.
	 * @param budgetCodes The account ledger codes connecting the budgets.
	 * @param banks The values for the banks.
	 * @param bankAccountTypes The account ledger codes connecting the banks.
	 * @param accountCodeNumberGenerator The generator for creating new account codes.
	 */
	BudgetBankLedgers(
		accounting::GeneralLedger ledgers,
		budgeting::Budget budget,
		foundation::TypeIndexMap<budgeting::BudgetSource, accounting::AccountCode> budgetCodes,
		BankMap banks,
		std::map<BankName, std::map<accounting::AccountNumber, BankAccountTypes>> bankAccountTypes,
		std::unique_ptr<accounting::AccountCodeNumberGenerator> accountCodeNumberGenerator
	);
	BudgetBankLedgers(const BudgetBankLedgers& other);
	BudgetBankLedgers(BudgetBankLedgers&& other) noexcept;
	~BudgetBankLedgers();
	BudgetBankLedgers& operator=(const BudgetBankLedgers& other) = delete;
	BudgetBankLedgers& operator=(BudgetBankLedgers&& other) noexcept;
	const BankLedgers& bankLedgers() const;
	const BudgetLedgers& budgetLedgers() const;
	/**
	 * @brief Changes the budgeting system objects being managed.
	 *
	 * @exception AccountError When there are issues with an account.
	 * @exception BankError When there are issues with a bank.
	 * @exception BudgetItemError When there are issues with a budget.
	 *
	 * @param ledgers The new general ledger books to start managing.
	 * @param budget The new budget to start managing.
	 * @param budgetCodes The associations between a budget item and the general ledger account for
	 * that item.
	 * @param banks The new financial institutions to start managing.
	 * @param bankAccountTypes The associations between financial institutions and the general
	 * ledger to start managing.
	 */
	void setManaged(
		accounting::GeneralLedger ledgers,
		budgeting::Budget budget,
		foundation::TypeIndexMap<budgeting::BudgetSource, accounting::AccountCode> budgetCodes,
		BankMap banks,
		std::map<
			budget_accounting::BankName,
			std::map<accounting::AccountNumber, budget_accounting::BankAccountTypes>
		> bankAccountTypes
	);
	/**
	 * @brief Determines if the object is in an empty state.
	 *
	 * @return Gives true when the banks, budget, and ledgers are all empty, other wise false.
	 */
	bool isEmpty() const;
	/**
	 * @brief Clears all the user data.
	 */
	void reset();
	/**
	 * @brief The current budget.
	 */
	const budgeting::Budget& budget() const;
	/**
	 * @brief The ledgers in the general ledger book.
	 */
	const accounting::GeneralLedger& ledgers() const;
	/**
	 * @brief The current banks.
	 */
	const BankMap& banks() const;
	/**
	 * @brief Generates an association between account codes, which are associated with a bank, and
	 * their user understandable bank account type and currency.
	 *
	 * @return The association of account codes to bank account types and its account currency held,
	 * or empty when no banks have any account associations.
	 */
	std::map<
		accounting::AccountCode,
		std::pair<budget_accounting::BankAccountTypes, pecunia::currency::Codes>
	> bankAccountTypeCurrency() const;
	/**
	 * @brief Changes the automated account code number generator.
	 *
	 * @pre The supplied generator cannot be null.
	 *
	 * @param accountCodeNumberGenerator The generator to change to.
	 */
	void setAccountCodeNumberGenerator(
		std::unique_ptr<accounting::AccountCodeNumberGenerator> accountCodeNumberGenerator
	);
	/**
	 * @brief An accessor to the automated account code number generator;
	 */
	const accounting::AccountCodeNumberGenerator& accountCodeNumberGenerator() const;
	/**
	 * @brief Collects all the known bank names.
	 */
	QSet<BankName> bankNames() const;

	//{ Budgeting Member Functions

	/**
	 * @brief Adds a new budget item to the budget and the associated ledger account.
	 *
	 * @exception BudgetItemError When the budget item source is the same as an existing account
	 * code name.
	 *
	 * @tparam BudgetItemType The type of the budget item to add.
	 *
	 * @param item The budget item that is to be added.
	 */
	template<typename BudgetItemType>
	void add(BudgetItemType item);
	template<typename BudgetItemType>
	void update(BudgetItemType item);
	void removeWage(const budgeting::BudgetSource& source);
	void removeBill(const budgeting::BudgetSource& source);
	void removeDebt(const budgeting::BudgetSource& source);
	void removeGoal(const budgeting::BudgetSource& source);
	void removeNontracked(const budgeting::BudgetSource& source);
	/**
	 * @brief Changes the name of a wage's source.
	 *
	 * @exception BudgetItemError When the budget item source is the same as an existing account
	 * code name.
	 *
	 * @param original The current budget item source.
	 * @param changed The new budget item source.
	 */
	void renameWage(
		const budgeting::BudgetSource& original,
		const budgeting::BudgetSource& changed
	);
	/**
	 * @brief Changes the name of a bill's source.
	 *
	 * @exception BudgetItemError When the budget item source is the same as an existing account
	 * code name.
	 *
	 * @param original The current budget item source.
	 * @param changed The new budget item source.
	 */
	void renameBill(
		const budgeting::BudgetSource& original,
		const budgeting::BudgetSource& changed
	);
	/**
	 * @brief Changes the name of a debt's source.
	 *
	 * @exception BudgetItemError When the budget item source is the same as an existing account
	 * code name.
	 *
	 * @param original The current budget item source.
	 * @param changed The new budget item source.
	 */
	void renameDebt(
		const budgeting::BudgetSource& original,
		const budgeting::BudgetSource& changed
	);
	/**
	 * @brief Changes the name of a goal's source.
	 *
	 * @exception BudgetItemError When the budget item source is the same as an existing account
	 * code name.
	 *
	 * @param original The current budget item source.
	 * @param changed The new budget item source.
	 */
	void renameGoal(
		const budgeting::BudgetSource& original,
		const budgeting::BudgetSource& changed
	);
	/**
	 * @brief Changes the name of a nontracked's source.
	 *
	 * @exception BudgetItemError When the budget item source is the same as an existing account
	 * code name.
	 *
	 * @param original The current budget item source.
	 * @param changed The new budget item source.
	 */
	void renameNontracked(
		const budgeting::BudgetSource& original,
		const budgeting::BudgetSource& changed
	);

	//}

	//{ Accounting Member Functions

	/**
	 * @brief Adds an account to the general ledger.
	 *
	 * @exception AccountError When the ledger contains an account code name that is same as a
	 * budget item source.
	 *
	 * @param account The account to add.
	 */
	void add(accounting::Account account);
	void close(const accounting::AccountCode& code);
	void open(const accounting::AccountCode& code);
	void remove(const accounting::AccountCode& code);
	/**
	 * @brief Recodes an account in the general ledger.
	 *
	 * @exception AccountError When the ledger contains an account code name that is same as a
	 * budget item source.
	 *
	 * @param original The current general ledger code.
	 * @param changed The new general ledger code.
	 */
	void recode(const accounting::AccountCode& original, const accounting::AccountCode& changed);
	void clear(const accounting::TransactionNumber& number);
	void reconcile(const accounting::TransactionNumber& number);
	accounting::TransactionNumber post(const accounting::Transaction& txn);
	void repost(const accounting::TransactionNumber& number, const accounting::Transaction& txn);
	void remove(const accounting::TransactionNumber& number);

	//}

	//{ Bank Member Functions

	/**
	 * @brief Adds a new bank to the system.
	 *
	 * @exception BankError When the financial institution is already closed.
	 * @exception BankError When the financial institution is already present in the container.
	 * @exception BankError When the ledger account type cannot be associated with the
	 * financial institution type.
	 *
	 * @param b The bank to add. This is expected not to have any accounts associated with it.
	 * @param accountNumbersBankTypes The general ledger account numbers and bank type associated
	 * with the bank.
	 */
	void add(
		Bank b,
		std::map<
			accounting::AccountNumber,
			budget_accounting::BankAccountTypes
		> accountNumbersBankTypes = {}
	);
	/**
	 * @brief Adds an association with an account code to a bank.
	 *
	 * @exception BankError When the bank does not exist.
	 * @exception BankError When another bank is already associated with that code.
	 * @exception BankError When the account code does not exist.
	 *
	 * @param name The name of the bank to add the account code to.
	 * @param number The number to add to the bank.
	 * @param type The type of bank account the general ledger account is managing.
	 */
	void addBankAccount(
		const BankName& name,
		const accounting::AccountNumber& number,
		const budget_accounting::BankAccountTypes& type
	);
	/**
	 * @brief Creates an account in the accounting system.
	 *
	 * @exception AccountError When the bank account type is not known.
	 * @exception AccountError When the account name is not valid.
	 * @exception AccountError When the account currency is not valid.
	 *
	 * @param name The name of the account to add.
	 * @param type The purpose type of the account.
	 * @param openingBalance The opening balance for the account.
	 *
	 * @return The code for the account created.
	 */
	accounting::AccountCode addAccount(
		std::string name,
		const budget_accounting::BankAccountTypes& type,
		const pecunia::currency::Money& openingBalance
	);
	/**
	 * @brief Updates the information for a financial institution and the associated accounts.
	 *
	 * @exception BankError When the financial institution is already closed.
	 * @exception BankError When the financial institution is not present in the container.
	 * @exception BankError When the ledger account is already associated with another financial
	 * institution.
	 * @exception BankError When the ledger account type cannot be associated with the
	 * financial institution type.
	 *
	 * @param b The financial institution whose information is to be updated.
	 * @param accountNumbersBankTypes The new account ledger associations for the financial
	 * institution.
	 */
	void update(
		Bank b,
		std::map<
			accounting::AccountNumber,
			budget_accounting::BankAccountTypes
		> accountNumbersBankTypes = {}
	);
	/**
	 * @brief Closes a bank and all associated financial institutions.
	 *
	 * @exception BankError When the financial institution is already closed.
	 * @exception BankError When the financial institution does not exist.
	 *
	 * @param name The name of the bank to close.
	 */
	void close(const BankName& name);
	/**
	 * @brief Opens a bank and all associated financial institutions.
	 *
	 * @exception BankError When the financial institution is already opened.
	 * @exception BankError When the financial institution does not exist.
	 *
	 * @param name The name of the bank to open.
	 */
	void open(const BankName& name);
	/**
	 * @brief Renames a bank and all associations under that name.
	 *
	 * @exception BankError When the financial institution name to be changed does not exist.
	 * @exception BankError When the financial institution name to be changed and to be changed
	 * into are the same.
	 * @exception BankError When the financial institution name to change to already exists.
	 * @exception BankError When the financial institution is already closed.
	 *
	 * @param original The current name of the bank.
	 * @param changed The new name for the bank.
	 */
	void rename(const BankName& original, const BankName& changed);
	/**
	 * @brief Removes a financial institution and all general ledgers under the financial
	 * institution's association.
	 *
	 * @exception BankError When the financial institution to remove does not exist.
	 * @exception BankError When the financial institution cannot be removed.
	 *
	 * @param bank The financial institution to remove.
	 */
	void remove(const BankName& bank);

	//}

	accounting::TransactionNumber post(
		const budgeting::BudgetItemDue& bid,
		const accounting::AccountCode& ac,
		const date::year_month_day& postOn
	);
	void skip(const budgeting::BudgetSource& bs);
};

DRN_BUDGET_ACCOUNTING_EXPORT bool operator==(const BudgetBankLedgers& lhs, const BudgetBankLedgers& rhs);
DRN_BUDGET_ACCOUNTING_EXPORT bool operator!=(const BudgetBankLedgers& lhs, const BudgetBankLedgers& rhs);
DRN_BUDGET_ACCOUNTING_EXPORT std::ostream& operator<<(std::ostream& o, const BudgetBankLedgers& bba);

}}

template<typename BudgetItemType>
void drn::budget_accounting::BudgetBankLedgers::add(BudgetItemType item)
{
	static_assert(
		std::is_base_of<drn::budgeting::BudgetedMoney, BudgetItemType>::value,
		"The budget item type must be derived from a drn::budgeting::BudgetedMoney."
	);
	assert(
		this->accountCodeNumberGenerator_ != nullptr
		&& "Account code number generator must be set."
	);
	this->verifyBudgetSourceName(item.source());
	this->budgetLedgers_.add(std::move(item), *this->accountCodeNumberGenerator_);
}

template<typename BudgetItemType>
void drn::budget_accounting::BudgetBankLedgers::update(BudgetItemType item)
{
	static_assert(
		std::is_base_of<drn::budgeting::BudgetedMoney, BudgetItemType>::value,
		"The budget item type must be derived from a drn::budgeting::BudgetedMoney."
	);
	assert(
		this->accountCodeNumberGenerator_ != nullptr
		&& "Account code number generator must be set."
	);
	this->budgetLedgers_.update(std::move(item), *this->accountCodeNumberGenerator_);
}

#endif
