/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020                                    ***
***  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_STRINGALGORITHM_H
#define DRN_FOUNDATION_STRINGALGORITHM_H

#include <sstream>
#include <string>
#include <utility>

#include "foundation_export.h"


namespace drn
{
namespace foundation
{
namespace string_algorithm
{

/**
 * @brief Removes all space characters on the left side of a string.
 *
 * @param[out] source The string whose white-space will be removed.
 */
DRN_FOUNDATION_EXPORT void ltrim(std::string& source);
/**
 * @brief Removes all space characters on the right side of a string.
 *
 * @param[out] source The string whose white-space will be removed.
 */
DRN_FOUNDATION_EXPORT void rtrim(std::string& source);
/**
 * @brief Removes all space characters on the left and right side of a string.
 *
 * @param[out] source The string whose white-space will be removed.
 */
DRN_FOUNDATION_EXPORT void trim(std::string& source);
/**
 * @brief Removes all space characters on the left and right side of a string.
 *
 * @param source The string whose white-space should be removed.
 *
 * @return The contents from the supplied source with the white-space removed.
 */
DRN_FOUNDATION_EXPORT std::string trim(const std::string& source);
/**
 * @brief Converts a value into its string representation.
 *
 * @tparam ValueType The type of the value to convert. This must support the stream insertion operator.
 *
 * @param value The value that is to be converted into a string.
 *
 * @return The value in string form.
 */
template<typename ValueType>
std::string asString(const ValueType& value);
/**
 * @brief Converts a pair of values into its string representation.
 *
 * @tparam FirstType The type of the first value.
 * @tparam SecondType The type of the second value
 *
 * @param value The paired value to convert.
 *
 * @return The first value in string form followed by the second value in string form separated by
 * the character sequence "=>".
 */
template<typename FirstType, typename SecondType>
std::string asString(const std::pair<FirstType, SecondType>& value);
/**
 * @brief Converts a container of values into its string representation.
 *
 * @tparam Container The type of container holding the values to convert.
 *
 * @param values The values to convert.
 * @param separator The separator to place between each value.
 *
 * @return The values in string form each separated by the separator and a space.
 */
template<typename Container>
std::string asString(const Container& values, const char separator);
/**
 * @brief Searches through the supplied string for the specified value.
 *
 * @param haystack The string to search through.
 * @param needle The character to locate.
 *
 * @return Gives true when the value is located, false else-wise.
 */
DRN_FOUNDATION_EXPORT bool contains(const std::string& haystack, const char& needle);
/**
 * @brief Searches through the supplied string for the specified value.
 *
 * @param haystack The string to search through.
 * @param needle The sequence of characters to locate.
 *
 * @return Gives true when the value is located, false else-wise.
 */
DRN_FOUNDATION_EXPORT bool contains(const std::string& haystack, const std::string& needle);

}}}

template<typename ValueType>
std::string drn::foundation::string_algorithm::asString(const ValueType& value)
{
	std::ostringstream oss{};
	oss << value;
	return oss.str();
}

template<typename FirstType, typename SecondType>
std::string drn::foundation::string_algorithm::asString(
	const std::pair<FirstType, SecondType>& value
)
{
	std::ostringstream oss{};
	oss << asString(value.first) << " => " << asString(value.second);
	return oss.str();
}

template<typename Container>
std::string drn::foundation::string_algorithm::asString(
	const Container& values,
	const char separator
)
{
	std::ostringstream oss{};
	bool isFirst{true};

	for (const auto& value : values)
	{
		oss << (isFirst ? std::string{""} : std::string{separator} + " ") << asString(value);
		isFirst = false;
	}
	return oss.str();
}

#endif
